""" A kernel transformation that precomputes quantities until a vector register
is filled and then does vector computations """

from dune.perftool.loopy.vcl import get_vcl_type_size
from dune.perftool.loopy.transformations.vectorview import (add_temporary_with_vector_view,
                                                            add_vector_view,
                                                            get_vector_view_name,
                                                            )
from dune.perftool.tools import get_pymbolic_basename

from loopy.kernel.creation import parse_domains
from loopy.symbolic import pw_aff_to_expr

from pymbolic.mapper.dependency import DependencyMapper
from pymbolic.mapper.substitutor import substitute

import pymbolic.primitives as prim
import loopy as lp
import numpy as np


def collect_vector_data_rotate(knl, insns, inames):
    #
    # Process/Assert/Standardize the input
    #

    # inames input -> tuple
    if isinstance(inames, str):
        inames = inames.split(",")
    inames = tuple(i.strip() for i in inames)

    # insns -> list of Instruction instances
    if isinstance(insns, lp.match.MatchExpressionBase):
        insns = lp.find_instructions(knl, insns)
    else:
        if isinstance(insns, str):
            insns = [i.strip() for i in insns.split(",")]
        insns = [knl.id_to_insn[i] for i in insns]

    # Analyse the inames of the given instructions and identify inames
    # that they all have in common. Those inames will also be iname dependencies
    # of inserted instructions.
    common_inames = frozenset([]).union(*(insn.within_inames for insn in insns)) - frozenset(inames)

    # Determine the vector lane width
    # TODO infer the numpy type here
    vec_size = get_vcl_type_size(np.float64)

    # Add an iname to the kernel which will be used for vectorization
    new_iname = "quad_vec_{}".format("_".join(inames))
    domain = "{{ [{0}] : 0<={0}<{1} }}".format(new_iname, str(vec_size))
    domain = parse_domains(domain, {})
    knl = knl.copy(domains=knl.domains + domain)
    knl = lp.tag_inames(knl, [(new_iname, "vec")])

    #
    # Inspect the given instructions for dependent quantities
    #

    quantities = {}
    for insn in insns:
        for expr in DependencyMapper()(insn.expression):
            basename = get_pymbolic_basename(expr)
            quantities.setdefault(basename, frozenset())
            quantities[basename] = quantities[basename].union(frozenset([expr]))
    assert all(len(q) == 1 for q in quantities.values())

    # Add vector size buffers for all these quantities
    replacemap_arr = {}
    replacemap_vec = {}
    for quantity in quantities:
        expr, = quantities[quantity]
        arrname = quantity + '_buffered_arr'
        knl = add_temporary_with_vector_view(knl,
                                             arrname,
                                             dtype=np.float64,
                                             shape=(vec_size,),
                                             dim_tags="c",
                                             base_storage=quantity + '_base_storage',
                                             )

        replacemap_arr[quantity] = prim.Subscript(prim.Variable(arrname), (prim.Variable('rotate_index'),))
        replacemap_vec[expr] = prim.Subscript(prim.Variable(get_vector_view_name(arrname)), (0, prim.Variable(new_iname),))

    write_match = lp.match.Or(tuple(lp.match.Writes(q) for q in quantities))
    iname_match = lp.match.And(tuple(lp.match.Iname(i) for i in inames))
    match = lp.match.And((write_match, iname_match))
    write_insns = lp.find_instructions(knl, match)

    other_insns = [i for i in knl.instructions if i.id not in [j.id for j in insns + write_insns]]
    new_insns = []
    temporaries = knl.temporary_variables

    for insn in write_insns:
        if isinstance(insn, lp.Assignment):
            new_insns.append(insn.copy(assignee=replacemap_arr[get_pymbolic_basename(insn.assignee)],
                                       )
                             )
        elif isinstance(insn, lp.CInstruction):
            pass
        else:
            raise NotImplementedError

    #
    # Add two counter variables to the kernel
    #

    # Insert a flat consecutive counter 'total_index'
    temporaries['total_index'] = lp.TemporaryVariable('total_index',  # name
                                                      dtype=np.int32,
                                                      )
    new_insns.append(lp.Assignment(prim.Variable("total_index"),  # assignee
                                   0,  # expression
                                   within_inames=common_inames,
                                   within_inames_is_final=True,
                                   id="assign_total_index",
                                   ))
    new_insns.append(lp.Assignment(prim.Variable("total_index"),  # assignee
                                   prim.Sum((prim.Variable("total_index"), 1)),  # expression
                                   within_inames=common_inames.union(inames),
                                   within_inames_is_final=True,
                                   depends_on=frozenset([i.id for i in write_insns]).union(frozenset({"assign_total_index"})),
                                   depends_on_is_final=True,
                                   id="update_total_index",
                                   ))

    # Insert a rotating index, that counts 0 , .. , vecsize - 1
    temporaries['rotate_index'] = lp.TemporaryVariable('rotate_index',  # name
                                                       dtype=np.int32,
                                                       )
    new_insns.append(lp.Assignment(prim.Variable("rotate_index"),  # assignee
                                   0,  # expression
                                   within_inames=common_inames,
                                   within_inames_is_final=True,
                                   id="assign_rotate_index",
                                   ))
    new_insns.append(lp.Assignment(prim.Variable("rotate_index"),  # assignee
                                   prim.Remainder(prim.Sum((prim.Variable("rotate_index"), 1)), vec_size),  # expression
                                   within_inames=common_inames.union(inames),
                                   within_inames_is_final=True,
                                   depends_on=frozenset([i.id for i in write_insns]).union(frozenset({"assign_rotate_index"})),
                                   depends_on_is_final=True,
                                   id="update_rotate_index",
                                   ))

    # Determine the condition for the continue statement
    upper_bound = prim.Product(tuple(pw_aff_to_expr(knl.get_iname_bounds(i).size) for i in inames))
    total_check = prim.Comparison(prim.Variable("total_index"), "<", upper_bound)
    rotate_check = prim.Comparison(prim.Variable("rotate_index"), "!=", 0)
    check = prim.LogicalAnd((rotate_check, total_check))

    # Insert the 'continue' statement
    new_insns.append(lp.CInstruction((),  # iname exprs that the code needs access to
                                     "continue;",  # the code
                                     predicates=frozenset({check}),
                                     depends_on=frozenset({"update_rotate_index", "update_total_index"}).union(frozenset([i.id for i in write_insns])),
                                     depends_on_is_final=True,
                                     within_inames=common_inames.union(inames),
                                     within_inames_is_final=True,
                                     id="continue_stmt",
                                     ))

    #
    # Reconstruct the compute instructions
    #

    for insn in insns:
        # Get a vector view of the lhs expression
        lhsname = get_pymbolic_basename(insn.assignee)
        knl = add_vector_view(knl, lhsname)
        lhsname = get_vector_view_name(lhsname)

        new_insns.append(lp.Assignment(prim.Subscript(prim.Variable(lhsname), (prim.FloorDiv(prim.Variable("total_size"), vec_size), prim.Variable(new_iname))),
                                       substitute(insn.expression, replacemap_vec),
                                       within_inames=frozenset(inames + (new_iname,)),
                                       within_inames_is_final=True,
                                       )
                         )

    return knl.copy(instructions=new_insns + other_insns)
